#pragma comment(linker, "/version:1.2") // MUST be in the form of major.minor

#include <windows.h>
#include <shlwapi.h>
#include <stdio.h>
#include <tchar.h>
#include "libs\SimpleString.h"

#ifndef READ_BUFFER_SIZE
#define READ_BUFFER_SIZE 0x10000
#endif

#define CREATION_ATTRIBUTE_MASK (FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_ARCHIVE | FILE_ATTRIBUTE_NORMAL | FILE_ATTRIBUTE_ENCRYPTED)

typedef struct {
	PVOID pvBuffer;          // file buffer (allocated only once)
	DWORD dwAttributesMask;  // mask of file attributes to preserve
	BOOL fVerbose;           // print status?
	BOOL fPreserveExisting;  // refuse to overwrite existing file?
	BOOL fDiscardStreams;    // discard NTFS streams?
	UINT cSuccessFiles;      // files copied
	UINT cSuccessDirs;       // directories processed
	UINT cErrors;            // errors encountered
	UINT cSkippedFiles;      // files skipped (if fPreserveExisting)
} WORKCTX, *PWORKCTX;

typedef struct {
	DWORD    dwFileAttributes;
	FILETIME ftCreationTime;
	FILETIME ftLastAccessTime;
	FILETIME ftLastWriteTime;
} FILEINFO, *PFILEINFO;

enum {
	TYPE_DNE = 0,
	TYPE_FILE,
	TYPE_DIR,
	TYPE_WILD
};

typedef union {
	UINT16 united;
	struct {
		UINT8 type;
		UINT8 fSlash;
	};
} ITEMTYPE, *PITEMTYPE;

VOID WINAPI CloneFile( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, PFILEINFO pfi );
VOID WINAPI CloneDirectory( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, PFILEINFO pfi );
VOID WINAPI CloneFilesIntoDirectory( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, BOOL fSourceIsDirectoryRoot );
UINT16 __fastcall DetermineItemType( PTSTR pszPath );
VOID __fastcall SetSparseAndCompression( HANDLE hFile, DWORD dwAttributes );
VOID __fastcall PrintError( PCTSTR pszPath, DWORD dwErrorCode );
__forceinline BOOL IsFlag( PCTSTR pszArg );
__forceinline BOOL CheckFlagI( PCTSTR pszArg, TCHAR ch );
__forceinline BOOL IsSpecialDirectoryName( PCTSTR pszPath );
__forceinline BOOL IsValidDirectory( PCTSTR pszPath );

UINT __fastcall EntryMain( UINT argc, PTSTR *argv )
{
	UINT iSourceStart = 0;
	UINT iSourceEnd;
	PTSTR pszTargetPath = NULL;
	BOOL fShowUsage = FALSE;

	// Parse command line parameters

	BOOL fFlagsDone = FALSE;
	UINT i;

	WORKCTX ctx;
	ZeroMemory(&ctx, sizeof(ctx));
	ctx.dwAttributesMask = ~0;

	for (i = 1; i < argc; ++i)
	{
		if (fFlagsDone || !IsFlag(argv[i]))
		{
			fFlagsDone = TRUE;

			if (iSourceStart == 0)
				iSourceStart = i;
			else
			{
				iSourceEnd = i;
				pszTargetPath = argv[i];
			}
		}
		else if (!ctx.fVerbose && CheckFlagI(argv[i], TEXT('v')))
			ctx.fVerbose = TRUE;
		else if (!ctx.fPreserveExisting && CheckFlagI(argv[i], TEXT('p')))
			ctx.fPreserveExisting = TRUE;
		else if (!ctx.fDiscardStreams && CheckFlagI(argv[i], TEXT('t')))
			ctx.fDiscardStreams = TRUE;
		else if (ctx.dwAttributesMask & FILE_ATTRIBUTE_READONLY && CheckFlagI(argv[i], TEXT('r')))
			ctx.dwAttributesMask &= ~FILE_ATTRIBUTE_READONLY;
		else if (ctx.dwAttributesMask & FILE_ATTRIBUTE_HIDDEN && CheckFlagI(argv[i], TEXT('h')))
			ctx.dwAttributesMask &= ~FILE_ATTRIBUTE_HIDDEN;
		else if (ctx.dwAttributesMask & FILE_ATTRIBUTE_SYSTEM && CheckFlagI(argv[i], TEXT('s')))
			ctx.dwAttributesMask &= ~FILE_ATTRIBUTE_SYSTEM;
		else if (ctx.dwAttributesMask & FILE_ATTRIBUTE_COMPRESSED && CheckFlagI(argv[i], TEXT('c')))
			ctx.dwAttributesMask &= ~FILE_ATTRIBUTE_COMPRESSED;
		else
			fShowUsage = TRUE;
	}

	// Act according to the command line parameters

	if (fShowUsage || pszTargetPath == NULL)
	{
		static const TCHAR szUsageTemplateGeneral[] = TEXT("  -%c  %s.\n");
		static const TCHAR szUsageTemplateAttribs[] = TEXT("  -%c  Strips away all %s file attributes.\n");

		_tprintf(TEXT("Usage: vcopy [-v] [-p] [-t] [-r] [-h] [-s] [-c] source [...] destination\n\n"));
		_tprintf(TEXT("Options:\n"));
		_tprintf(szUsageTemplateGeneral, TEXT('v'), TEXT("Verbose mode; prints status"));
		_tprintf(szUsageTemplateGeneral, TEXT('p'), TEXT("Preserves destination files; do not overwrite files"));
		_tprintf(szUsageTemplateGeneral, TEXT('t'), TEXT("Strips away streams and other NTFS-specific metadata"));
		_tprintf(szUsageTemplateAttribs, TEXT('r'), TEXT("read-only"));
		_tprintf(szUsageTemplateAttribs, TEXT('h'), TEXT("hidden"));
		_tprintf(szUsageTemplateAttribs, TEXT('s'), TEXT("system"));
		_tprintf(szUsageTemplateAttribs, TEXT('c'), TEXT("compressed"));

		return(~1);
	}
	else
	{
		/* Source->Destination behaviors:
		 * - wild->file|DNE  ERROR
		 * - wild->dir\|dir  destdir\*
		 * - file->file|DNE  destfile
		 * - file->dir\|dir  destdir\file
		 * - dir->file       ERROR
		 * - dir->dir\       destdir\dir
		 * - dir->dir|DNE    destdir
		 *
		 * Additional rules:
		 * - A destination of dir\ must be a valid existing directory
		 * - A non-wildcard source must exist
		 * - A destination may not contain a wildcard
		 */

		BOOL fMustCopyIntoTarget;

		ITEMTYPE itTarget;
		itTarget.united = DetermineItemType(pszTargetPath);

		// If the user explicitly tacked on a slash to the end of the target or
		// if there are multiple source items, then the target must be a
		// directory that is copied into, not onto.

		fMustCopyIntoTarget = itTarget.fSlash || (iSourceEnd - iSourceStart > 1);

		if ( (itTarget.type != TYPE_DIR && fMustCopyIntoTarget) ||
		     (itTarget.type == TYPE_WILD) )
		{
			++ctx.cErrors;
			PrintError(pszTargetPath, (itTarget.type == TYPE_DNE) ? ERROR_PATH_NOT_FOUND : ERROR_BAD_PATHNAME);
			iSourceEnd = 0;
		}

		// Allocate a file buffer if we need to copy manually
		if (ctx.fDiscardStreams && NULL == (ctx.pvBuffer = VirtualAlloc(NULL, READ_BUFFER_SIZE, MEM_COMMIT, PAGE_READWRITE)))
			return(~0);

		for (i = iSourceStart; i < iSourceEnd; ++i)
		{
			PTSTR pszSourcePath = argv[i];

			ITEMTYPE itSource;
			itSource.united = DetermineItemType(pszSourcePath);

			if ( (itSource.type == TYPE_WILD && itTarget.type != TYPE_DIR) ||
			     (itSource.type == TYPE_DIR && itTarget.type == TYPE_FILE) )
			{
				// This will never be tripped if there are multiple sources
				++ctx.cErrors;
				PrintError(pszTargetPath, (itTarget.type == TYPE_DNE) ? ERROR_PATH_NOT_FOUND : ERROR_BAD_PATHNAME);
				continue;
			}
			else if (itSource.type == TYPE_DNE)
			{
				++ctx.cErrors;
				PrintError(pszSourcePath, (itSource.fSlash) ? ERROR_PATH_NOT_FOUND : ERROR_FILE_NOT_FOUND);
				continue;
			}
			else if (itSource.type != TYPE_DIR && itSource.fSlash)
			{
				++ctx.cErrors;
				PrintError(pszSourcePath, ERROR_DIRECTORY);
				continue;
			}

			// Dispatch the copy operation
			if (itSource.type == TYPE_FILE && itTarget.type != TYPE_DIR)
				CloneFile(&ctx, pszSourcePath, pszTargetPath, NULL);
			else if (itSource.type == TYPE_DIR && !fMustCopyIntoTarget)
				CloneDirectory(&ctx, pszSourcePath, pszTargetPath, NULL);
			else
				CloneFilesIntoDirectory(&ctx, pszSourcePath, pszTargetPath, FALSE);
		}

		// Cleanup
		if (ctx.pvBuffer) VirtualFree(ctx.pvBuffer, 0, MEM_RELEASE);

		// Post-mortem
		_tprintf(TEXT("\t%d file(s) and %d folder(s) copied"), ctx.cSuccessFiles, ctx.cSuccessDirs);
		if (ctx.cSkippedFiles) _tprintf(TEXT(", %d file(s) skipped"), ctx.cSkippedFiles);
		if (ctx.cErrors) _tprintf(TEXT(", with %d error(s)"), ctx.cErrors);
		_tprintf(TEXT(".\n"));
	}

	return(ctx.cErrors);
}

VOID WINAPI CloneFile( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, PFILEINFO pfi )
{
	BOOL fSuccess = FALSE;
	BY_HANDLE_FILE_INFORMATION bhfi;

	if (pctx->fVerbose)
		_tprintf(TEXT("%s\n"), pszSourcePath);

	/* CopyFile does more than just copy: it has special handling of reparse
	 * points, security descriptors, and NTFS streams, so until that degree of
	 * completeness can be duplicated, the manual copy can be used only when
	 * such NTFS-specific information is to be discarded; future implementation
	 * of these features will enable finer control and better efficiency.
	 */

	if (pctx->fDiscardStreams)
	{
		/* Manual copy */

		HANDLE hFileIn, hFileOut;

		hFileIn = CreateFile(
			pszSourcePath,
			GENERIC_READ,
			FILE_SHARE_READ,
			NULL,
			OPEN_EXISTING,
			FILE_FLAG_SEQUENTIAL_SCAN,
			NULL
		);

		if (hFileIn != INVALID_HANDLE_VALUE)
		{
			if (GetFileInformationByHandle(hFileIn, &bhfi))
			{
				if (pfi == NULL)
					pfi = (PFILEINFO)&bhfi;

				pfi->dwFileAttributes &= pctx->dwAttributesMask;

				// There is no need to check if the source and target files are
				// the same because the request of write access to a file that
				// does not have write sharing will fail.

				hFileOut = CreateFile(
					pszTargetPath,
					GENERIC_READ | GENERIC_WRITE,
					FILE_SHARE_READ,
					NULL,
					(pctx->fPreserveExisting) ? CREATE_NEW : CREATE_ALWAYS,
					FILE_FLAG_SEQUENTIAL_SCAN | (pfi->dwFileAttributes & CREATION_ATTRIBUTE_MASK),
					hFileIn
				);

				if (hFileOut != INVALID_HANDLE_VALUE)
				{
					DWORD cbRead, cbWritten;
					ULONGLONG cbTotalRead = 0, cbTotalWritten = 0;

					ULARGE_INTEGER cbFileSize;
					cbFileSize.LowPart = bhfi.nFileSizeLow;
					cbFileSize.HighPart = bhfi.nFileSizeHigh;

					// The sparse/compression attributes need to be set prior
					// to writing any data.

					SetSparseAndCompression(hFileOut, pfi->dwFileAttributes);

					// The actual data copy...

					do
					{
						ReadFile(hFileIn, pctx->pvBuffer, READ_BUFFER_SIZE, &cbRead, NULL);

						if (cbRead)
						{
							WriteFile(hFileOut, pctx->pvBuffer, cbRead, &cbWritten, NULL);
							cbTotalRead += cbRead;
							cbTotalWritten += cbWritten;
						}

					} while (cbRead == READ_BUFFER_SIZE);

					// Sanity check and time set

					if ( cbFileSize.QuadPart == cbTotalRead &&
					     cbFileSize.QuadPart == cbTotalWritten )
					{
						fSuccess = TRUE;

						SetFileTime(
							hFileOut,
							&pfi->ftCreationTime,
							&pfi->ftLastAccessTime,
							&pfi->ftLastWriteTime
						);
					}

					CloseHandle(hFileOut);

					if (fSuccess)
						SetFileAttributes(pszTargetPath, pfi->dwFileAttributes);

				} // if valid destination handle

			} // if valid file information

			CloseHandle(hFileIn);
		}
	}
	else
	{
		/* Using CopyFile */

		if (pfi == NULL)
		{
			HANDLE hFile = CreateFile(
				pszSourcePath,
				0,
				FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
				NULL,
				OPEN_EXISTING,
				0,
				NULL
			);

			if (hFile != INVALID_HANDLE_VALUE)
			{
				if (GetFileInformationByHandle(hFile, &bhfi))
					pfi = (PFILEINFO)&bhfi;

				CloseHandle(hFile);
			}
		}

		if (CopyFile(pszSourcePath, pszTargetPath, pctx->fPreserveExisting))
			fSuccess = TRUE;

		if (fSuccess && pfi)
		{
			HANDLE hFile;

			// Turn off the read-only bit so that we can modify the file

			DWORD dwAttributes = GetFileAttributes(pszTargetPath);
			if (dwAttributes != INVALID_FILE_ATTRIBUTES && dwAttributes & FILE_ATTRIBUTE_READONLY)
				SetFileAttributes(pszTargetPath, dwAttributes & ~FILE_ATTRIBUTE_READONLY);

			pfi->dwFileAttributes &= pctx->dwAttributesMask;

			hFile = CreateFile(
				pszTargetPath,
				GENERIC_READ | GENERIC_WRITE,
				FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
				NULL,
				OPEN_EXISTING,
				0,
				NULL
			);

			if (hFile != INVALID_HANDLE_VALUE)
			{
				SetSparseAndCompression(hFile, pfi->dwFileAttributes);

				SetFileTime(
					hFile,
					&pfi->ftCreationTime,
					&pfi->ftLastAccessTime,
					&pfi->ftLastWriteTime
				);

				CloseHandle(hFile);
			}

			SetFileAttributes(pszTargetPath, pfi->dwFileAttributes);
		}
	}

	if (fSuccess)
	{
		++pctx->cSuccessFiles;
	}
	else
	{
		DWORD dwErrorCode = GetLastError();

		if (dwErrorCode == ERROR_FILE_EXISTS && pctx->fPreserveExisting)
		{
			++pctx->cSkippedFiles;
			if (pctx->fVerbose)
				_ftprintf(stderr, TEXT("Already exists: %s\n"), pszTargetPath);
		}
		else
		{
			++pctx->cErrors;
			PrintError(pszSourcePath, dwErrorCode);
		}
	}
}

VOID WINAPI CloneDirectory( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, PFILEINFO pfi )
{
	BY_HANDLE_FILE_INFORMATION bhfi;

	if (pctx->fVerbose)
		_tprintf(TEXT("%s\n"), pszSourcePath);

	// Step 1: Create the target directory

	(pctx->fDiscardStreams) ?
		CreateDirectory(pszTargetPath, NULL) :
		CreateDirectoryEx(pszSourcePath, pszTargetPath, NULL);

	// Step 2: Clone the attributes

	if (IsValidDirectory(pszTargetPath))
	{
		// Ensure that we have the source attribute information

		if (pfi == NULL)
		{
			HANDLE hFile = CreateFile(
				pszSourcePath,
				0,
				FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
				NULL,
				OPEN_EXISTING,
				FILE_FLAG_BACKUP_SEMANTICS,
				NULL
			);

			if (hFile != INVALID_HANDLE_VALUE)
			{
				if (GetFileInformationByHandle(hFile, &bhfi))
					pfi = (PFILEINFO)&bhfi;

				CloseHandle(hFile);
			}
		}

		// If the folder is compressed or encrypted, we should set the right
		// bits BEFORE copying any files...

		if (pfi)
		{
			pfi->dwFileAttributes &= pctx->dwAttributesMask;
			SetFileAttributes(pszTargetPath, pfi->dwFileAttributes);

			if (pfi->dwFileAttributes & FILE_ATTRIBUTE_COMPRESSED)
			{
				HANDLE hFile = CreateFile(
					pszTargetPath,
					GENERIC_READ | GENERIC_WRITE,
					FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
					NULL,
					OPEN_EXISTING,
					FILE_FLAG_BACKUP_SEMANTICS,
					NULL
				);

				if (hFile != INVALID_HANDLE_VALUE)
				{
					SetSparseAndCompression(hFile, FILE_ATTRIBUTE_COMPRESSED);
					CloseHandle(hFile);
				}
			}
			else if (pfi->dwFileAttributes & FILE_ATTRIBUTE_ENCRYPTED)
			{
				EncryptFile(pszTargetPath);
			}
		}
	}
	else
	{
		++pctx->cErrors;
		PrintError(pszTargetPath, GetLastError());
		return;
	}

	// Step 3: Clone the content

	CloneFilesIntoDirectory(pctx, pszSourcePath, pszTargetPath, TRUE);

	// Step 4: Clone the timestamp

	if (pfi)
	{
		HANDLE hFile = CreateFile(
			pszTargetPath,
			GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
			NULL,
			OPEN_EXISTING,
			FILE_FLAG_BACKUP_SEMANTICS,
			NULL
		);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			SetFileTime(
				hFile,
				&pfi->ftCreationTime,
				&pfi->ftLastAccessTime,
				&pfi->ftLastWriteTime
			);

			CloseHandle(hFile);
		}
	}

	++pctx->cSuccessDirs;
}

VOID WINAPI CloneFilesIntoDirectory( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, BOOL fSourceIsDirectoryRoot )
{
	HANDLE hFind;
	WIN32_FIND_DATA finddata;

	// Copy the paths to a scratchpad buffer
	TCHAR szSourcePath[MAX_PATH << 2];
	TCHAR szTargetPath[MAX_PATH << 2];
	PTSTR pszSourcePathAppend = SSChainNCpy(szSourcePath, pszSourcePath, SSLen(pszSourcePath));
	PTSTR pszTargetPathAppend = SSChainNCpy(szTargetPath, pszTargetPath, SSLen(pszTargetPath));

	// Ensure that we have a slash termination for the target path
	*pszTargetPathAppend++ = TEXT('\\');

	// Prep parameters for FindFirstFile and the establish the append position
	*pszSourcePathAppend = 0;
	if (fSourceIsDirectoryRoot)
		SSCpy4Ch(pszSourcePathAppend++, TEXT('\\'), TEXT('*'), 0, 0);
	else if (pszSourcePathAppend = StrRChr(szSourcePath, pszSourcePathAppend, TEXT('\\')))
		++pszSourcePathAppend;
	else
		pszSourcePathAppend = szSourcePath;

	if ((hFind = FindFirstFile(szSourcePath, &finddata)) != INVALID_HANDLE_VALUE)
	{
		do
		{
			if (!IsSpecialDirectoryName(finddata.cFileName))
			{
				SSCpy(pszSourcePathAppend, finddata.cFileName);
				SSCpy(pszTargetPathAppend, finddata.cFileName);

				if (finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
					CloneDirectory(pctx, szSourcePath, szTargetPath, (PFILEINFO)&finddata);
				else
					CloneFile(pctx, szSourcePath, szTargetPath, (PFILEINFO)&finddata);
			}

		} while (FindNextFile(hFind, &finddata));

		FindClose(hFind);
	}
}

UINT16 __fastcall DetermineItemType( PTSTR pszPath )
{
	PTSTR psz = pszPath;

	ITEMTYPE it;
	it.united = 0;

	while (*psz)
	{
		switch (*psz)
		{
			case TEXT('*'):
			case TEXT('?'):
				it.type = TYPE_WILD;
				break;

			case TEXT('/'):
				*psz = TEXT('\\');

			// Any encounter with a slash following a wildcard must be invalid,
			// and we flag that by setting fSlash with TYPE_WILD.
			case TEXT('\\'):
				if (it.type == TYPE_WILD)
					it.fSlash = TRUE;
				break;
		}

		++psz;
	}

	while (--psz >= pszPath && *psz == TEXT('\\'))
	{
		*psz = 0;
		it.fSlash = TRUE;
	}

	if (it.fSlash && *pszPath == 0)
	{
		// Special case: "\"
		it.type = TYPE_DIR;
	}
	else if (it.type == TYPE_DNE)
	{
		DWORD dwAttributes = GetFileAttributes(pszPath);

		if (dwAttributes != INVALID_FILE_ATTRIBUTES)
			it.type = (dwAttributes & FILE_ATTRIBUTE_DIRECTORY) ? TYPE_DIR : TYPE_FILE;
	}

	return(it.united);
}

VOID __fastcall SetSparseAndCompression( HANDLE hFile, DWORD dwAttributes )
{
	DWORD cbRet;

	if (dwAttributes & FILE_ATTRIBUTE_SPARSE_FILE)
	{
		FILE_SET_SPARSE_BUFFER buf = { TRUE };
		DeviceIoControl(hFile, FSCTL_SET_SPARSE, &buf, sizeof(buf), NULL, 0, &cbRet, NULL);
	}

	if (dwAttributes & FILE_ATTRIBUTE_COMPRESSED)
	{
		USHORT buf = COMPRESSION_FORMAT_DEFAULT;
		DeviceIoControl(hFile, FSCTL_SET_COMPRESSION, &buf, sizeof(buf), NULL, 0, &cbRet, NULL);
	}
}

VOID __fastcall PrintError( PCTSTR pszPath, DWORD dwErrorCode )
{
	_ftprintf(stderr, TEXT("Error: %s"), pszPath);

	if (dwErrorCode)
	{
		PTSTR pszErrorMessage = NULL;

		DWORD cchMessage = FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			dwErrorCode,
			0,
			(PTSTR)&pszErrorMessage,
			0,
			NULL
		);

		if (cchMessage)
		{
			pszErrorMessage[StrCSpn(pszErrorMessage, TEXT("\r\n"))] = 0;
			_ftprintf(stderr, TEXT(" [%s]"), pszErrorMessage);
		}
		else
		{
			_ftprintf(stderr, TEXT(" [%08X]"), dwErrorCode);
		}

		LocalFree(pszErrorMessage);
	}

	_ftprintf(stderr, TEXT("\n"));
}

__forceinline BOOL IsFlag( PCTSTR pszArg )
{
	return(
		(pszArg[0] | 0x02) == TEXT('/') &&
		(pszArg[1]       ) != 0 &&
		(pszArg[2]       ) == 0
	);
}

__forceinline BOOL CheckFlagI( PCTSTR pszArg, TCHAR ch )
{
	return(
		(pszArg[1] | 0x20) == ch
	);
}

__forceinline BOOL IsSpecialDirectoryName( PCTSTR pszPath )
{
	// TRUE if name is "." or ".."

	#ifdef UNICODE
	return(
		(*((UPDWORD)pszPath) == WCHARS2DWORD(L'.', 0)) ||
		(*((UPDWORD)pszPath) == WCHARS2DWORD(L'.', L'.') && pszPath[2] == 0)
	);
	#else
	return(
		(*((UPWORD)pszPath) == CHARS2WORD('.', 0)) ||
		(*((UPWORD)pszPath) == CHARS2WORD('.', '.') && pszPath[2] == 0)
	);
	#endif
}

__forceinline BOOL IsValidDirectory( PCTSTR pszPath )
{
	DWORD dwAttributes = GetFileAttributes(pszPath);
	return(dwAttributes != INVALID_FILE_ATTRIBUTES && dwAttributes & FILE_ATTRIBUTE_DIRECTORY);
}
